/*
 * example_atomic.c
 */
#include <linux/atomic.h>
#include <linux/bitops.h>
#include <linux/module.h>
#include <linux/printk.h>

#define BYTE_TO_BINARY_PATTERN "%c%c%c%c%c%c%c%c"
#define BYTE_TO_BINARY(byte)                                                   \
    ((byte & 0x80) ? '1' : '0'), ((byte & 0x40) ? '1' : '0'),                  \
        ((byte & 0x20) ? '1' : '0'), ((byte & 0x10) ? '1' : '0'),              \
        ((byte & 0x08) ? '1' : '0'), ((byte & 0x04) ? '1' : '0'),              \
        ((byte & 0x02) ? '1' : '0'), ((byte & 0x01) ? '1' : '0')

/* 演示原子加法和减法操作 */
static void atomic_add_subtract(void)
{
    atomic_t rimian;
    atomic_t xunmi = ATOMIC_INIT(50);
	/* 赋值为45 */
    atomic_set(&rimian, 45);
    /* 减去1(45-1=44) */
    atomic_dec(&rimian);
	/* 加上7(44+7=51) */
    atomic_add(7, &rimian);
    /* 加1(51+1=52) */
    atomic_inc(&rimian);
	/* 两个名字, 没有其他意义 */
    pr_info("寻觅: %d, 日冕: %d\n", atomic_read(&xunmi),
            atomic_read(&rimian));
}

/* 演示原子位操作 */
static void atomic_bitwise(void)
{
    unsigned long word = 0;

    pr_info("初始 0: " BYTE_TO_BINARY_PATTERN, BYTE_TO_BINARY(word));
    set_bit(3, &word);    // 设置第3位为1
    set_bit(5, &word);    // 设置第5位为1
    pr_info("修正 1: " BYTE_TO_BINARY_PATTERN, BYTE_TO_BINARY(word));
    clear_bit(5, &word);  // 清除第5位
    pr_info("修正 2: " BYTE_TO_BINARY_PATTERN, BYTE_TO_BINARY(word));
    change_bit(3, &word); // 切换第3位的值
    pr_info("修正 3: " BYTE_TO_BINARY_PATTERN, BYTE_TO_BINARY(word));

    // 检查第3位，如果原本为1，则返回 `非预期值` 的输出, 如果不是1则将其设置为1
	if (test_and_set_bit(3, &word))   
			pr_info("非预期值\n");
    pr_info("修正 4: " BYTE_TO_BINARY_PATTERN, BYTE_TO_BINARY(word));

    word = 255;  // 设置word的所有位为1
    pr_info("修正 5: " BYTE_TO_BINARY_PATTERN "\n", BYTE_TO_BINARY(word));
}

static int __init example_atomic_init(void)
{
    pr_info("原子操作示例开始\n");

    atomic_add_subtract();
    atomic_bitwise();

    return 0;
}

static void __exit example_atomic_exit(void)
{
    pr_info("原子操作示例退出\n");
}

module_init(example_atomic_init);
module_exit(example_atomic_exit);

MODULE_DESCRIPTION("原子操作示例");
MODULE_LICENSE("GPL");
