/*
 * example_rwlock.c
 */
#include <linux/module.h>
#include <linux/printk.h>
#include <linux/rwlock.h>

static DEFINE_RWLOCK(myrwlock);

static void example_read_lock(void)
{
    unsigned long flags;

    read_lock_irqsave(&myrwlock, flags);
    pr_info("读锁已加锁\n");

    /* 从某处读取数据 */

    read_unlock_irqrestore(&myrwlock, flags);
    pr_info("读锁已解锁\n");
}

static void example_write_lock(void)
{
    unsigned long flags;

    write_lock_irqsave(&myrwlock, flags);
    pr_info("写锁已加锁\n");

    /* 向某处写入数据 */

    write_unlock_irqrestore(&myrwlock, flags);
    pr_info("写锁已解锁\n");
}

static int __init example_rwlock_init(void)
{
    pr_info("example_rwlock 模块启动\n");

    example_read_lock();
    example_write_lock();

    return 0;
}

static void __exit example_rwlock_exit(void)
{
    pr_info("example_rwlock 模块退出\n");
}

module_init(example_rwlock_init);
module_exit(example_rwlock_exit);

MODULE_DESCRIPTION("读/写锁示例");
MODULE_LICENSE("GPL");
