/*
 * example_tasklet.c
 */
#include <linux/delay.h>
#include <linux/interrupt.h>
#include <linux/module.h>
#include <linux/printk.h>

/* 宏 DECLARE_TASKLET_OLD 存在是为了兼容性 参见 https://lwn.net/Articles/830964/ */
#ifndef DECLARE_TASKLET_OLD
#define DECLARE_TASKLET_OLD(arg1, arg2) DECLARE_TASKLET(arg1, arg2, 0L)
#endif

/* tasklet_fn 是 tasklet 执行的函数，接收一个 unsigned long 类型的参数 */
static void tasklet_fn(unsigned long data)
{
    pr_info("示例 tasklet 开始执行!\n");
    mdelay(20000);  // 延迟 10 秒
    pr_info("示例 tasklet 执行结束!\n");
}

/* 定义一个 tasklet，并将 tasklet_fn 作为其处理函数 */
static DECLARE_TASKLET_OLD(mytask, tasklet_fn);

/* 模块初始化函数 */
static int __init example_tasklet_init(void)
{
    pr_info("tasklet 示例初始化\n"); 
    tasklet_schedule(&mytask);  // 安排 tasklet 执行
    mdelay(10);
    pr_info("示例 tasklet 初始化继续...\n"); 
    return 0;
}

/* 模块退出函数 */
static void __exit example_tasklet_exit(void)
{
    pr_info("tasklet 示例退出\n"); 
    tasklet_kill(&mytask);  // 终止 tasklet 的执行
}

module_init(example_tasklet_init);
module_exit(example_tasklet_exit); 

MODULE_DESCRIPTION("Tasklet 示例");
MODULE_LICENSE("GPL");

